/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   *
 *     as this notice is not removed.                                     *
 **************************************************************************/
 
/*
	collect.c
*/

/*
 *  CHANGES:
 *    27Jan89 Merit: Added check for max number of hosts
 *    13Oct89 Merit: Added new parameters:
 *
 *       -m mode
 *      Set file mode (octal) of log files on creation.  Handy if you don't
 *      want to mess with umask.
 *
 *       -u
 *      Write logfile times in Universal Time instead of local.
 *
 *      Rel 2.4:
 *   30Nov89 Merit/ISI: New parameter for terse mode.
 *       -t
 *      Decrease volume of log file. 
 *
 */
#include <stdio.h>
#include <rpc/rpc.h>
#include <errno.h>
#include <ctype.h>
#include <strings.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/file.h>
#include <netdb.h>

#include "collect.h"

char *HELPSTR = "Usage:\n\
  collect object(s)\
\n\t [-e enumFile]\
\n\t [-h host1] .. [-h hostn]\
\n\t [-i min] \tPolling interval\
\n\t [-c min] \tCheckpoint interval\
\n\t [-p port] \tTCP port to connect to\
\n\t [-r min] \tClear interval\
\n\t [-m mode] \tSet log file mode\
\n\t [-u] \t\tFormat times in universal time\
\n\t [-d] \t\tTrace & log to stdout\
\n\t [-dl] \t\tTrace to stdout, log to files\
\n\t [-dx] \t\tTrace & hex dump to stdout\n";

int	interval,	/* Minutes between queries */
	chkpnt,		/* Minutes between checkpoints */
	clear;		/* Minutes between clears */
	
struct collector rspy[MAX_ADDRS];	

FILE *tracefp = NULL;		/* File to save trace */
boolean hexd = FALSE,
	needlog = TRUE;
extern int port;
extern boolean print_utime;
extern boolean tersemode;
extern int openmode;
char out[256];

void hex_dump();
void netclose();
void flush_xdr();
void closeconn();

XDR read_s, write_s;
XDR *readp = &read_s, *writep = &write_s;

#define error_out(x,y) {fprintf(stderr,"collect: %s %s\n",x,y); exit(1);}

void
Usage()
{
   printf(HELPSTR);
   exit(1);
}

char *argscan(argcp, argvp)
   int *argcp;
   char ***argvp;
{
   char *cp;
	
   if (*(cp = 2+**argvp)) 
      return(cp);
   else if  (--*argcp > 0)
      return(*++*argvp);
   Usage();
} /* argscan */

main(argc,argv)
    int argc;
    char *argv[];
{
   int 	i, j, h, 
   	host = 0,	/* index for collector structure */
   	times;		/* loop control */
   u_long temp;
   long now, nextread;
   char *cp, *number;
   FILE *fp;
   struct tm *tp;
   
   tracefp = NULL;

   if (argc == 1) 
      Usage();

   /* Initialization stuff */
      now = time(0);
      tp = localtime(&now);
      /* nextread = creation -(creation mod 60 min) */
      rspy[0].creation = now;
      nextread = now - ((tp->tm_min*60) + tp->tm_sec);

      while (--argc > 0 && *++argv) {
	 if (!strncmp("-e", *argv, 2)) {
	     cp = argscan(&argc, &argv);
	     if ((fp = fopen(cp, "r")) != NULL) {
                scan_enum(fp);
		if (tracefp) {
		   show_enums(out, sizeof(out));
		   fprintf(tracefp, "%s\n", out);
		   }
		fclose(fp);
	        }
	     else {
                perror("collect: Cannot open parm file") ;
                exit(1) ;
	        }
	     }
	 else if (!strncmp("-h", *argv, 2)) {
	    if (host >= MAX_ADDRS - 1) {
		perror("collect: Too many hosts");
		exit(1);
	    }
	    cp = argscan(&argc, &argv);
	    if (resolve_name(cp, &rspy[host].inetaddr, 1) < 1) {
	       error_out("Bad hostname:", cp);
	       }
	    else {
	       strcpy(rspy[host++].hostname, cp);
	       if (tracefp) {
		  temp = ntohl(rspy[host-1].inetaddr.s_addr);
	          fprintf(tracefp,"Resolved: %s = %d.%d.%d.%d\n", 
		     rspy[host-1].hostname,
		     (temp>>24), ((temp>>16)&0xff), 
		     ((temp>>8)&0xff), (temp&0xff));
		  }
	       }
	    }
	 else if (!strcmp("-d", *argv)) {
	     /* trace & log info to stdout */
	     needlog = FALSE;
	     tracefp = stdout;
	     }
	 else if (!strcmp("-dl", *argv)) {
	     /* trace info to stdout, log info to files */
	     tracefp = stdout;
	     }
	 else if (!strcmp("-dx", *argv)) {
	     /* trace to stdout, log is hex dump to stdout */
	     tracefp = stdout;
	     hexd = TRUE;
	     needlog = FALSE;
	     }
	 else if (!strncmp("-i", *argv, 2)) {
	     number = argscan(&argc, &argv);
	     for (cp = number; *cp; cp++) 
	        if (! isdigit(*cp)) {
		   error_out("Bad polling interval:", number);
		   }
	     interval = atoi(number);
	     }
	 else if (!strncmp("-c", *argv, 2)) {
	     number = argscan(&argc, &argv);
	     for (cp = number; *cp; cp++) 
	        if (! isdigit(*cp)) {
		   error_out("Bad checkpoint interval:", number);
		   }
	     chkpnt = atoi(number);
	     }
	 else if (!strncmp("-r", *argv, 2)) {
	     number = argscan(&argc, &argv);
	     for (cp = number; *cp; cp++) 
	        if (! isdigit(*cp)) {
		   error_out("Bad clear interval:", number);
		   }
	     clear = atoi(number);
	     }
	 else if (!strncmp("-m", *argv, 2)) {
	     /* set log file mode */
	     number = argscan(&argc, &argv);
	     if (sscanf(number, "%o", &openmode) != 1)
		error_out("Bad open mode:", number);
	     }
	 else if (!strcmp("-u", *argv)) {
	     /* print times in universal time */
	     print_utime = TRUE;
	     }
     else if (!strcmp("-t", *argv)) {
	      /* generate log in terse mode */
          tersemode = TRUE;
	     }
	 else if (!strncmp("-p", *argv, 2)) {
	     number = argscan(&argc, &argv);
	     if (!(port = get_port(number,"collect")))
	        return;
	     }
	 else if (**argv == '-') {
	    fprintf(stderr, "collect: Bad option: %s\n", *argv);
	    Usage();
	    }
	 else { /* copy to first blank slot for all hosts */
	    for (j=0; j<MAX_STATS; j++) 
	       if (!rspy[0].s[j].query) {
		  for (h=0; h < MAX_ADDRS; h++) {
		     strncpy(rspy[h].s[j].sname, *argv, 
		     		sizeof(rspy[h].s[j].sname)-1);
		     rspy[h].s[j].query = TRUE;
		     }
		  break;
	          }
	    }
         }
	 
   if (!rspy[0].s[0].query) {
      fprintf(stderr, "collect: No object specified!\n");
      Usage();
      }
      
   if (! host) {
      /* Default host name */
      gethostname(rspy[0].hostname, sizeof(rspy[0].hostname));
      if (!resolve_name(rspy[0].hostname, &rspy[0].inetaddr, 1))
         error_out("Bad default host name:", rspy[0].hostname);
      }
      
   for (host = 0; strlen(rspy[host].hostname); host++) {
	rspy[host].creation = rspy[0].creation;
	rspy[host].nextclear = (clear)? nextread: 0;
	}
   if ((clear || chkpnt) && (!interval))
      interval = ((chkpnt<clear) && chkpnt)? chkpnt: clear;
      
   for (times = (interval)? 0: 1; (times || interval); times--) {
      for (host = 0; strlen(rspy[host].hostname); host++) {
	 if (tracefp) fprintf(tracefp,"\nCollector: %s\n",rspy[host].hostname);
	 /* open tcp connection */
	 if (openconn(&rspy[host].inetaddr)) {

	    /* go through collection structures */
	    for (i = 0; rspy[host].s[i].query ; i++) {
	       rspy[host].readtime = time(0);
	       signal(SIGALRM, netclose); /* reinit timer for each read */
	       alarm(TIMEOUT);
	       if (! read_stats(host, i))
	          break;
	       if (hexd)
	          hex_dump(&read_s, tracefp, out, sizeof(out), 0, 
		  		DTYP_bits);
	       else
	          log_reply(&rspy[host]);
	       if (clear) {
		  while (rspy[host].readtime >= rspy[host].nextclear)
		     rspy[host].nextclear += (60*clear);
		  }
	       }
	  
	    signal(SIGALRM, NULL);	/* clear timeout */
	    alarm(0);
	    closeconn(); 
	    }
	 if (tracefp) fflush(tracefp);
	 }
      if (interval) {
	 now = time(0);
	 while (now >= nextread) nextread += (60*interval);
         sleep((nextread - now));
	 }
      } /* keep going */

	 
} /* main */

boolean
read_stats(host, index)
   int host, index;
{
   char *cp, temp[200];
   int i;

   sprintf(temp, "%s ;", rspy[host].s[index].sname);
   cp = temp;
   i = (clear && (rspy[host].readtime >= rspy[host].nextclear))? 
       READCL_op: READ_op;

   if (xdr_int(&write_s, &i) &&
       xdr_string(&write_s, &cp, strlen(temp))) {
      if (tracefp) 
         fprintf(tracefp,"\n%s '%s'\n", 
	 		 	(i == READCL_op)? "Read+Clear": "Read", 
				rspy[host].s[index].sname);
      flush_xdr();
      return(TRUE);
      }
   return(FALSE);
} /* read_stats */

/*
int
gobble(count)
  int count;
{
   int i,j;	
   for (j=0;j<count;j++) {
     xdr_int(&read_s, &i);
     if (tracefp) fprintf(tracefp," %d",i);
     }
   if (tracefp) fprintf(tracefp,"\n");
   return(i);
} 
*/
