/*
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static  char rcsid[] =
    "@(#)$Header: etherif-bpf.c,v 1.4 90/11/05 15:17:02 vern Locked $ (LBL)";
#endif

#include <stdio.h>
#include <netdb.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <net/bpf.h>
#include <net/if.h>

#include "interface.h"

static void
bpf_stats(fd)
	int fd;
{
	struct bpf_stat s;
	
	if (ioctl(fd, BIOCGSTATS, &s) < 0)
		return;

	fflush(stdout);
	(void)fprintf(stderr, "\n%d packets received by filter\n", s.bs_recv);
	(void)fprintf(stderr, "%d packets dropped by kernel\n", s.bs_drop);
}

void
readloop(cnt, if_fd, fp, printit)
	int cnt;
	int if_fd;
	struct bpf_program *fp;
	void (*printit)();
{
	u_char *buf;
	u_int bufsize;

	buf = readinit(if_fd, fp, &bufsize);
	while (1) {
		cnt = readether(cnt, if_fd, printit, buf, bufsize, cflag);
		if (cflag && cnt == 0)
			break;
	}
	wrapup(if_fd);
}

u_char *
readinit(if_fd, fp, bufsize_addr)
	int if_fd;
	struct bpf_program *fp;
	u_int *bufsize_addr;
{
	u_char *buf;
	u_int bufsize;

	if (ioctl(if_fd, BIOCGBLEN, (caddr_t)&bufsize) < 0) {
		perror("getting buffer length");
		exit(1);
	}
	buf = (u_char *)malloc(bufsize);

	if (fp) {
		if (ioctl(if_fd, BIOCSETF, (caddr_t)fp) < 0) {
			perror("setting filter");
			exit(1);
		}
	}

	*bufsize_addr = bufsize;
	return(buf);
}

int
readether(cnt, if_fd, printit, buf, bufsize, cntflag)
	int cnt;
	int if_fd;
	void (*printit)();
	u_char *buf;
	u_int bufsize;
{
	register u_char *bp;
	struct bpf_hdr *bhp;
	int datalen, inc;
	int cc;
	static long file_offset = 0;

	if ((cc = read(if_fd, (char *)buf, bufsize)) < 0) {
		perror("bpf read");
		abort();
		exit(1);
	}

	file_offset += cc;
	if (file_offset > (1 << 30)) {
		/* Oops - we're in danger of wrapping the kernel file offset
		 * for /dev/bpf if we read another 2^30 bytes.  Do a seek()
		 * to reset the file offset.  Don't complain if the seek()
		 * returns an error, as on some systems we may not be allowed
		 * to seek on /dev/bpf (but hopefully on those systems, no
		 * error is generated when the kernel offset wraps!).
		 */
		(void) lseek(if_fd, 0L, 0);
		file_offset = 0;
	}

	/*
	 * Loop through each packet.
	 */
	bp = buf;
	while (cc > 0) {
		bhp = (struct bpf_hdr *)bp;
		datalen = bhp->bh_datalen;
		(*printit)(bp + bhp->bh_hdrlen, &bhp->bh_tstamp,
			   datalen);
		if (cntflag && --cnt == 0)
			break;
		if (datalen > snaplen)
			datalen = snaplen;
		inc = BPF_WORDALIGN(datalen + bhp->bh_hdrlen);
		cc -= inc;
		bp += inc;
	}

	return(cnt);
}

void wrapup(fd)
	int fd;
{
	bpf_stats(fd);
	close(fd);
}

static inline int
bpf_open()
{
	int fd;
	int n = 0;
	char device[sizeof "/dev/bpf000"];

	/*
	 * Go through all the minors and find one that isn't in use. 
	 */
	do {
		sprintf(device, "/dev/bpf%d", n++);
		fd = open(device, O_RDONLY);
	} while (fd < 0 && errno == EBUSY);

	if (fd < 0) {
		perror(device);
		exit(-1);
	}
	return fd;
}

int
initdevice(device, pflag, linkp)
	char *device;
	int pflag;
	struct dlinfo *linkp;
{
	struct timeval timeout;
	int if_fd;
	int i;
	struct ifreq ifr;
	struct bpf_devp devp;
	
	if_fd = bpf_open();

	strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));
	if (ioctl(if_fd, BIOCSETIF, &ifr) < 0) {
		perror(device);
		exit(-1);
	}
	/* Get the data link layer type. */
	if (ioctl(if_fd, BIOCDEVP, &devp) < 0) {
		perror(device);
		exit(-1);
	}
	linkp->dl_type = devp.bdev_type;
	linkp->off_netlayer = devp.bdev_hdrlen;
	linkinfo_to_typeoff(linkp);

	/* set timeout */
	timeout.tv_sec = 1;
	timeout.tv_usec = 0;
	if (ioctl(if_fd, BIOCSRTIMEOUT, &timeout) < 0) {
		perror(device);
		exit(-1);
	}
	/* set promiscuous mode if requested, but not for SLIP */
	if (pflag == 0 && devp.bdev_type != DLT_SLIP) {
		if (ioctl(if_fd, BIOCPROMISC, (void *)0) < 0) {
			perror(device);
			exit(-1);
		}
	}
	/* set truncation */
	if (ioctl(if_fd, BIOCTRUNCATE, &snaplen) < 0) {
		perror(device);
		exit(-1);
	}
	/* Snapshot length of data portion. */
	snapdlen = snaplen - devp.bdev_hdrlen;

	return(if_fd);
}
