/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   *
 *     as this notice is not removed.                                     *
 **************************************************************************/
 
/* CHANGES:
 *   17Nov88 RTB: Add EQ object
 */
 
/*
 *   Internet Statistics Facility -- Generic Operation Routines
 *
 *    This module defines filter objects, which can examine given values
 *    and make a choice of alternative sets of calls on recording Statistical
 *    Objects.
 *
 *    These filters are constructed as abstract objects, to parallel
 *    the form of the recording Statistical Objects.  They keep simple
 *    counts of total packets and true choices (the difference being the
 *    false-choice count).
 */
 
#include <stdio.h>
#include <sys/types.h>
#include "stat.h"
#include "sobj.h"
 
/*
 *
 *          RF Object Class ==  RANGE-CHECK FILTER 
 *
 *    Objects of the RF class are filters that check that the given field
 *    value falls outside a specified range.
 *
 *    Operations on RF class objects:
 *
 *       SOBJ *Make_RF( objname, &{L, U}, nparms, datalen )
 *
 *           If nparms == 1, take it as L == U
 *
 *       int Write_RF( (SOBJ *) SOBJ-ptr, &value , length)
 *
 *           => True if field value falls inside range:
 *                  L <= x <= U.
 *
 *       Read_RF( fd, (SOBJ *) SOBJ-ptr )
 *
 *       Clear_RF(  (SOBJ *) SOBJ-ptr )
 *
 *       Delete_RF(  (SOBJ *) SOBJ-ptr )
 *
 *
 */
    

SOBJ *Make_RF() ;
int  Write_RF(), Read_RF();
void Clear_RF(), Delete_RF() ;

        /*   Transfer Vector 
         *   for external access to these routines
         */
GENERICOP RFop = { Make_RF, Write_RF, Read_RF, Clear_RF, Delete_RF } ;

#define RFA struct stat_farea
RFA {
    SOBJ  rfa_sobj ;
        
    long  rfa_trueno ;
    union v_union rfa_ULvals;  /* L, U values.  These are stored like parms,
                                *   right-justified in fullword for L <= 4,
                                *   else left-justified.
                                */
} ;

#define rfa_L  rfa_ULvals.v_ul.v_ul1
#define rfa_U  rfa_ULvals.v_ul.v_ul3
#define rfa_Lb rfa_ULvals.v_byte[0]
#define rfa_Ub rfa_ULvals.v_byte[MAX_DLENG]

struct RF_parms {
    u_long rfp_val[LONGPERVAL];
};

 
SOBJ *Make_RF(objname, parmp, nparms, datalen)
    char *objname;
    struct RF_parms *parmp ;
    int datalen, nparms;
    {
    register RFA *rfap = (RFA *) malloc(sizeof(RFA));
    int len = ALIGN(datalen);
    u_char *pp = (u_char *) parmp;
    
    if (rfap == NULL) {
        SOBJ_error = "RF: memory shortage";
        return(NULL) ;
    } 
    else if (nparms == 0 || nparms > 2) {
        SOBJ_error = "RF: wrong # parms";
        return(NULL);
    }
    
    Clear_RF(rfap);
    bcopy(pp, &rfap->rfa_Lb, len);
    bcopy(pp+len, (nparms == 1)? &rfap->rfa_Lb:&rfap->rfa_Ub, len);
    return((SOBJ *) rfap) ;
}


int Write_RF(rfap, Valuep, L)
    RFA  *rfap ;
    char *Valuep ;
    register int L;
    { 
    register int result; 
    
    if (L > sizeof(long))
        result = (strncmp(Valuep, &rfap->rfa_Lb, L) >= 0 &&
                  strncmp(Valuep, &rfap->rfa_Ub, L) <= 0);
    else {
        u_long Value = LFETCH( Valuep, L) ; 
        result = (Value >= rfap->rfa_L && Value <= rfap->rfa_U);
    }                   
    if (result) rfap->rfa_trueno++;
    return((result)? WRRET_TRUE : WRRET_FALSE ) ; 
}

void Clear_RF(rfap)
    RFA  *rfap ;
    {
    rfap->rfa_trueno = 0 ;
}


void Delete_RF(rfap)
    RFA  *rfap ;
    {
    free(rfap) ;
}


int Read_RF(fd, rfap)
    int fd;
    RFA *rfap ;
    {                     
    if (!PutReadResp(fd, RFr_SIZE, rfap) ) return(FALSE);
    
    if (ISTTY(fd))
        printf("  TRUE no= %d for range: (%d, %d)\n", 
            rfap->rfa_trueno, rfap->rfa_L, rfap->rfa_U);
    else {
        if (!xdr_int(fd, &rfap->rfa_trueno)) return(FALSE) ;  
    }
    return(TRUE);     
}
    
  
/*
 *
 *          SF Object Class ==  SET-CHECK FILTER 
 *
 *    Objects of the SF class filter by checking whether a given field
 *    value is a member of a specified set of values.
 *
 *    SF objects are implemented using a precomputed open hashing.  When
 *    the table is computed it tries to choose a hash table large enough
 *    to get "perfect" hashing. 
 *
 *
 *    Operations on SF class objects:
 *
 *       SOBJ *Make_SF( object-name, &valuelist[#bins], #values, datalen  )
 *
 *       int Write_SF( (SOBJ *) SOBJ-ptr, &value , length)
 *             Returns TRUE if value falls into given set.
 *
 *       Read_SF( fd, (SOBJ *) SOBJ-ptr )
 *
 *       Delete_SF( (SOBJ *) SOBJ-ptr )
 *
 *       Clear_SF( (SOBJ *) SOBJ-ptr )
 *
 *    THIS CLASS USES THE ROUTINES OF CLASS FO (FREQ-ONLY).
 */
    
extern SOBJ *Make_FO() ;
extern boolean  Search_FO();
extern u_long BinCount_FO();
boolean Read_SF();
void  Clear_FO(), Delete_FO() ;

    /* Transfer Vector 
     *   for external access to these routines
     */
GENERICOP SFop = { Make_FO, Search_FO, Read_SF, Clear_FO, Delete_FO } ;


int Read_SF(fd, sfap)
    int fd;
    SOBJ *sfap;
    {   
    u_long trueno = BinCount_FO(sfap) ;
    
    if (!PutReadResp(fd, SFr_SIZE, sfap) ) return(FALSE);
    
    if (ISTTY(fd))
        printf("  TRUE no= %d\n", trueno);
    else {
        if (!xdr_int(fd, &trueno)) return(FALSE) ;  
    }
    return(TRUE);                     
}

    
    
  
/*
 *
 *          EQ Object Class ==  EQUALITY FILTER 
 *
 *    Objects of the EQ class filter by checking whether a given field
 *    value is equal to a specified value.
 * *
 *
 *    Operations on EQ class objects:
 *
 *       SOBJ *Make_EQ( object-name, &value, 1, datalen  )
 *
 *       int Write_EQ( (SOBJ *) SOBJ-ptr, &value , length)
 *             Returns TRUE if value falls into given set.
 *
 *       Read_EQ( fd, (SOBJ *) SOBJ-ptr )
 *
 *       Delete_EQ( (SOBJ *) SOBJ-ptr )
 *
 *       Clear_EQ( (SOBJ *) SOBJ-ptr )
 *
 */
     

SOBJ *Make_EQ() ;
int  Write_EQ(), Read_EQ();
void Clear_EQ(), Delete_EQ() ;

        /*   Transfer Vector 
         *   for external access to these routines
         */
GENERICOP EQop = { Make_EQ, Write_EQ, Read_EQ, Clear_EQ, Delete_EQ } ;

#define EQA struct stat_eqarea
EQA {
    SOBJ  eqa_sobj ;
        
    long  eqa_trueno ;
    char  eqa_val[MAX_DLENG];  /* Comparison Value */
} ;

struct EQ_parms {
    u_long eqp_val[LONGPERVAL];
};

 
SOBJ *Make_EQ(objname, parmp, nparms, datalen)
    char *objname;
    struct EQ_parms *parmp ;
    int datalen, nparms;
    {
    EQA *eqap;
    int offset =  PARMALIGN(datalen);
    int len = ALIGN(datalen);
    u_char *pp = (u_char *) parmp + offset;
    
    if ((eqap = (EQA *) malloc(sizeof(EQA))) == NULL) {
        SOBJ_error = "EQ: memory shortage";
        return(NULL) ;
    } 
    else if (nparms == 0 || nparms > 2) {
        SOBJ_error = "EQ: wrong # parms";
        return(NULL);
    }
    
    Clear_EQ(eqap);
    bcopy(pp, eqap->eqa_val, len);
    return((SOBJ *) eqap) ;
}


int Write_EQ(eqap, Valuep, L)
    EQA  *eqap ;
    register char *Valuep ;
    register int L;
    { 
    register char *cp = &eqap->eqa_val[0]; 
    
    while (L-- > 0) 
    	if (*Valuep++ != *cp++) return(WRRET_FALSE);

    eqap->eqa_trueno++;
    return(WRRET_TRUE) ; 
}

void Clear_EQ(eqap)
    EQA  *eqap ;
    {
    eqap->eqa_trueno = 0 ;
}


void Delete_EQ(eqap)
    EQA  *eqap ;
    {
    free(eqap) ;
}


int Read_EQ(fd, eqap)
    int fd;
    EQA *eqap ;
    {                     
    if (!PutReadResp(fd, SFr_SIZE, eqap) ) return(FALSE);
    
    if (ISTTY(fd))
        printf("  TRUE no= %d\n", eqap->eqa_trueno);
    else {
        if (!xdr_int(fd, &eqap->eqa_trueno)) return(FALSE) ;  
    }
    return(TRUE);     
}
