
 

/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   * 
 *     as this notice is not removed.                                     *
 **************************************************************************/ 

/*                   parse.c
 *     PARSE PACKETS COLLECTED FROM ETHERNET
 */
 
/* CHANGES:
 *    18Nov88 ISI: Time preamble more efficient, and remove unneeded rlen.
 *                       INTERVAL field conditionally compiled.
 *    19Nov88 ISI: Add reference count and cumulative invocation count to FPT.
 *    22Nov88 ISI: Add Ethernet Filter Table 
 *    15Oct89 ISI: Changes for Sun4 (SPARC) -- with thanks to C.Philip Wood(LANL)
 *    17Oct89 AKS/RTB: Changes for subnets
 *          Rel 2.4:
 *     1Nov89 ISI: Change handling of packet length -- new field "length"
 *     9Nov89 RTB: Add symmetric-other field to support symif statement
 */
  
#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include "stat.h" 
#include "sobj.h"
#include "packet.h"

#ifdef BPF
#include "interface.h"

extern char *EtherIface;
int if_fd = -1;

extern u_char *bpf_buf;
extern u_int bpf_bufsize;
u_char *bpf_buf;
u_int bpf_bufsize;

deinitdevice()
{
	if (if_fd >= 0)
		wrapup(if_fd);
}
#endif	/* BPF */
    
#define Ether_hdr_ptr ((struct ether_header *) pp)

    /*        Fields
     *
     *  NOTE: THESE MUST BE IN PARSE ORDER... i.e., as a packet is
     *  parsed, the sequence of defined field numbers MUST! be MONTONIC.
     *
     *  Compilation flags:
     *    #define FALIGN:  Include code for SPARC chip; harmless but inefficient
     *               for Sun3 hardware.
     *    #define SUBNETS: Include subnet fields SSBNET and DSBNET
     *    #define IBMRTPC: Modify parse for Network Tap header of PC RT
     *    #define vax:     Little-endian byte ordering
     *    #define XPARSE:  Include experimental virtual field INTERVAL
     *
     */
#define FI_ROOT        1
#define FI_PACKET      1  /* pseudo-field ... actually, entire packet! */
#define FI_LENGTH      2  /* Length of packet (excl. of Ethernet hdr) */
#define FI_INTERVAL    3  /* Time in ms since previous packet */
#define FI_PERSEC      4  /* Count of packets received in last second */
#define FI_ETHER_SRC   5
#define FI_ETHER_DST   6
#define FI_ETHER_TYPE  7
#define FI_IP_VERSION  8  /*IP version: Defined only if NOT current version */
#define FI_IP_PKTLEN   9  /* (This is really obsolete, with 'length' field) */
#define FI_IP_OPTION   10 /*IP option code: defined only if there is option*/
#define FI_IP_TOS      11
#define FI_IP_OFFSET   12 /*IP reas'bly offset: Defined only if pkt fragm'ted*/
#define FI_IP_PROT     13
#define FI_IP_SRCNET   14 /* Network of source address */
#define FI_IP_DSTNET   15 /* Network of dest address */
#ifdef SUBNETS
#define FI_IP_SSBNET   16 /* Subnetwork of source address */
#define FI_IP_DSBNET   17 /* Subnetwork dest address */
#define FI_IP_SRCHST   18
#define FI_IP_DSTHST   19
#define FI_TCP_SPORT   20
#define FI_TCP_DPORT   21
#define FI_UDP_SPORT   22
#define FI_UDP_DPORT   23
#define FI_ICMP_TYPE   24
#define MAXFI_NO 24
#else  /* No subnets */
#define FI_IP_SRCHST   16
#define FI_IP_DSTHST   17
#define FI_TCP_SPORT   18
#define FI_TCP_DPORT   19
#define FI_UDP_SPORT   20
#define FI_UDP_DPORT   21
#define FI_ICMP_TYPE   22
#define MAXFI_NO 22
#endif  /* SUBNETS */

  /* Virtual fields... */
static int zero = 0;
static u_long snet, dnet, ssbnet, dsbnet;
static u_char ip_version;
static short  ip_offset;
static long   interval, secrun;
       short  packetlen;  /* length of Ethernet data ...
                           * THIS VAR. IS ACCESSED DIRECTLY BY -bytes OBJECTS
                           */
Subnet SubHdr, *SubTail = &SubHdr;  /* Dummy Subnet entry to head
   subnet table, and ptr to last entry in list */

#define FIELD_DEF(num, parent, so, name, ptr, len, type) \
 num, parent, so, 0, name, NULL, (char *)ptr, len, len, type, 0 , 0   

/*            PACKET PARSE TREE:
 *
 *packet
 *  length
 *    Ether.src
 *      Ether.dst
 *        Ether.type
 *          IP.version
 *   ----- Stop parse here if IP.version != 4 ...
 *            IP.length  (Same value as "length", but only for IP packets)
 *              IP.option
 *                IP.TOS
 *                  IP.offset
 *   ----- Stop parse here if offset != 0 (not first fragment) ...
 *                    IP.protocol
 *                      IP.srcnet
 *                        IP.dstnet
 *                          IP.ssubnet
 *                            IP.dsubnet
 *                              IP.srchost
 *                                IP.dsthost
 *                                  TCP.srcport
 *                                    TCP.dstport
 *                                  UDP.srcport
 *                                    UDP.dstport
 *                                  ICMP.type
 *
 */     
FIELD Fields[MAXFI_NO+2] = {
 {NULL},  /* unused */      
 {FIELD_DEF(FI_PACKET, 0, 0, "packet", NULL, MAX_BPSIZE, DTYP_packet)},
 {FIELD_DEF(FI_LENGTH, FI_PACKET, 0, "length", &packetlen, 2, DTYP_int)},
 {FIELD_DEF(FI_INTERVAL, -1, 0, "interval.ms", &interval, 4, DTYP_int)},
 {FIELD_DEF(FI_PERSEC, -1, 0, "per.sec", &secrun, 4, DTYP_int)},
 
 {FIELD_DEF(FI_ETHER_SRC, FI_LENGTH, FI_ETHER_DST,
                                        "Ether.src", NULL, 6, DTYP_EtherA)},
 {FIELD_DEF(FI_ETHER_DST, FI_ETHER_SRC, FI_ETHER_SRC, 
                                        "Ether.dst", NULL, 6, DTYP_EtherA)},
 {FIELD_DEF(FI_ETHER_TYPE, FI_ETHER_DST, 0, 
                                        "Ether.type", NULL, 2, DTYP_bits)},
 {FIELD_DEF(FI_IP_VERSION, FI_ETHER_TYPE, 0, 
                                        "IP.version", &ip_version, 1, DTYP_int)},
 {FIELD_DEF(FI_IP_PKTLEN, FI_IP_VERSION, 0, 
                                        "IP.length", &packetlen, 2, DTYP_int)},

 {FIELD_DEF(FI_IP_OPTION, FI_IP_PKTLEN, 0, "IP.option", NULL, 1, DTYP_int)},
 {FIELD_DEF(FI_IP_TOS, FI_IP_OPTION, 0, "IP.TOS", NULL, 1, DTYP_bits)},
 {FIELD_DEF(FI_IP_OFFSET, FI_IP_TOS, 0, "IP.offset", &ip_offset, 2, DTYP_int)},
 {FIELD_DEF(FI_IP_PROT, FI_IP_OFFSET, 0, "IP.protocol", NULL, 1, DTYP_int)},
 {FIELD_DEF(FI_IP_SRCNET, FI_IP_PROT, FI_IP_DSTNET, 
                                        "IP.srcnet", &snet, 4, DTYP_IP)},
 {FIELD_DEF(FI_IP_DSTNET, FI_IP_SRCNET, FI_IP_SRCNET, 
                                        "IP.dstnet",  &dnet, 4, DTYP_IP)},
#ifdef SUBNETS
 {FIELD_DEF(FI_IP_SSBNET, FI_IP_DSTNET, FI_IP_DSBNET, 
                                        "IP.srcsubn", &ssbnet, 4, DTYP_IP)},
 {FIELD_DEF(FI_IP_DSBNET, FI_IP_SSBNET, FI_IP_SSBNET, 
                                        "IP.dstsubn", &dsbnet, 4, DTYP_IP)},
 {FIELD_DEF(FI_IP_SRCHST, FI_IP_DSBNET, FI_IP_DSTHST, 
                                        "IP.srchost", NULL, 4, DTYP_IP)},
#else
 {FIELD_DEF(FI_IP_SRCHST, FI_IP_DSTNET, FI_IP_DSTHST, 
                                        "IP.srchost", NULL, 4, DTYP_IP)},
#endif
 {FIELD_DEF(FI_IP_DSTHST, FI_IP_SRCHST, FI_IP_SRCHST, 
                                        "IP.dsthost", NULL, 4, DTYP_IP)},
 {FIELD_DEF(FI_TCP_SPORT, FI_IP_DSTHST, FI_TCP_DPORT, 
                                        "TCP.srcport", NULL, 2, DTYP_int)},
 {FIELD_DEF(FI_TCP_DPORT, FI_TCP_SPORT, FI_TCP_SPORT, 
                                        "TCP.dstport", NULL, 2, DTYP_int)},
 {FIELD_DEF(FI_UDP_SPORT, FI_IP_DSTHST, FI_UDP_DPORT,
                                        "UDP.srcport", NULL, 2, DTYP_int)},
 {FIELD_DEF(FI_UDP_DPORT, FI_UDP_SPORT, FI_UDP_SPORT, 
                                        "UDP.dstport", NULL, 2, DTYP_int)},
 {FIELD_DEF(FI_ICMP_TYPE, FI_IP_DSTHST, 0, "ICMP.type", NULL, 1, DTYP_int)},
 {NULL}  /* END OF FIELDS */
};

/* Ethernet Filter Table.  Has one entry for each Ethernet type value
 *  that the parse subroutine defines.
 */
struct FilterTE EtherFTable[] =
    { {FI_IP_VERSION, 0x0800}, 
      {0, 0}  /* flag end of list */
};
   
    
struct timeval last_time = {0, 0};
            
int    packetno;   /* Max # of original packets to receive (0 => no limit)*/
u_long packetcnt;  /* Number received */

u_long StartSecCnt= 0;  /* Packet count at start of second interval */
u_long StartTickCnt= 0; /* Packet count at start of tick interval */
u_long MaxSecRun = 0;   /* Max # packets arriving within one second */
u_long MaxTickRun = 0;  /* Max # packets arriving within one clock tick */
   
#ifdef vax  
    /* Network mask in network byte order... */
long netmask[4] = {0x000000FF, 0x000000FF, 0x0000FFFF,  0x00FFFFFF} ;  
#else           
    /* Network byte order == host byte order */
long netmask[4] = {0xFF000000, 0xFF000000, 0xFFFF0000,  0xFFFFFF00} ; 
#endif          

#define INTERPRET(FIno, p)  { Fields[FIno].field_ptr= (char *) p;  \
                              if (Fields[FIno].field_opcs) \
                                 interpret(Fields[FIno].field_opcs);  }  

#define INTERPRETV(FIno, p)  {  if (Fields[FIno].field_opcs) \
                                 interpret(Fields[FIno].field_opcs);  }
         /* Version of INTERPRET for virtual-fields */  


    /*
     *
     *  PACKET PARSER
     *
     */   
parse( pp, clen, rlen, capture_time)
    register u_char *pp ;
    int clen ;   /* amount of data captured */
    int rlen ;   /* actual packet length */
    struct timeval capture_time;
    {
    register struct ip *ip_h;
    register u_char *optp;         
    register union ip_data *ipdp;     
    register x; 
    union {  /* place to full-word align two fullword IP addresses*/
        char ip_addrs_c[8];
        struct {
            struct in_addr saddr; /* src addr */
            struct in_addr daddr; /* dst addr */
        } ip_addrs_s;
    } ip_addrs;
#define SADDR ip_addrs.ip_addrs_s.saddr.s_addr 
#define DADDR ip_addrs.ip_addrs_s.daddr.s_addr 
    Subnet *FindSubnet();
        
    packetcnt++;
    packetlen = rlen - sizeof (struct ether_header);
    if (packetno && packetcnt >= packetno) exit(0); 
               /* For overhead measurements */
    
#ifdef IBMRTPC
    /*  RT packets have the network tap header on the front (which
        provides fullword alignment for the IP, TCP, UDP headers). */
    clen -= sizeof(struct tap_header);
    pp += sizeof(struct tap_header);
#endif

    Fields[FI_PACKET].field_len = clen ;  /* save len captured */
    INTERPRET(FI_PACKET, (char *) pp);
     
    CurrTime =  capture_time.tv_sec;  /* Time NIT captured the packet */

#ifdef XPARSE   
       /* INTERVAL: special field, interval in ms; cannot be combined with
        *   any later field (would have no meaning).  INTERVAL was defined
        *   to gather histogram of interarrival intervals; not generally
        *   useful.
        */ 
    if (Fields[FI_INTERVAL].field_opcs) { 
            /* If object attached to INTERVAL psuedo-field... */
    	interval =  (capture_time.tv_sec - last_time.tv_sec)*1000 + 
                (capture_time.tv_usec - last_time.tv_usec)/1000 ;
    	interpret(Fields[FI_INTERVAL].field_opcs);
    } 
#endif 
  
    if (capture_time.tv_usec != last_time.tv_usec) {
            /*** This assumes microsec clock advances in discrete steps of
            /***    one clock tick (20 ms on Sun 3).
            /*** If new clock tick... ***/
        if (MaxTickRun < (x = packetcnt - StartTickCnt)) 
       		MaxTickRun = x;
        StartTickCnt = packetcnt;
    }
    
    if (capture_time.tv_sec != last_time.tv_sec) {
            /*** Count # packets arriving in same second ***/
        if (MaxSecRun < (secrun = packetcnt - StartSecCnt))
            MaxSecRun = secrun;
        INTERPRETV(FI_PERSEC, &secrun);
        StartSecCnt = packetcnt;
    } 
    last_time = capture_time;
    
        /* BEGIN THE PACKET PARSING... */
    
    INTERPRET(FI_LENGTH, &packetlen) ;
    INTERPRET(FI_ETHER_SRC, &Ether_hdr_ptr->ether_shost) ;
    INTERPRET(FI_ETHER_DST, &Ether_hdr_ptr->ether_dhost) ;
    INTERPRET(FI_ETHER_TYPE, &Ether_hdr_ptr->ether_type) ;

    switch (Ether_hdr_ptr->ether_type) {
    
/*    Statistics on ARP or RARP contents unimplemented...
 *
 *  case ETHERTYPE_ARP:
 *  case ETHERTYPE_REVARP:
 *      break ;
 */
            
    case ETHERTYPE_IP:
        /********************** IP datagram *************************/
        
        ip_h = (struct ip *)(pp+sizeof(struct ether_header));
            
            /* Now that we have discovered that Butterfly gateways use IP
             * Version 3 for inter-gateway encapsulation, we are not interested
             * in the version number.  But need to test it, to avoid
             * illegal parses.
             *
             * Note that IP_VERSION pseudo-field is defined ONLY if the 
             * version is not 4.
             */
        ip_version = ip_h->ip_v ;                
        if (ip_version != IPVERSION)  {             
            INTERPRETV(FI_IP_VERSION, &ip_version) ;    
            return;                                 
        }   
            /*
             *  Must test for fragments... assume all headers will fit into
             *   first fragment, just ignore later fragments.  Note that the
             *   IP length field that is recorded is the fragment length.
             *   There is no simple way to get the true packet length; would
             *   essentially have to build reassembly mechanism to match
             *   first and last fragments of each packet.
             */
        
        if (ip_offset = ip_h->ip_off)   
            if ((ip_offset = ((ip_h->ip_off&0x1fff)<<3)) ||
                                       (ip_h->ip_off&IP_MF)) {
                    /* 
                     * This is a fragment... define IP-OFFSET pseudo-field 
                     */
                INTERPRETV(FI_IP_OFFSET, &ip_offset);
                if (ip_offset) return;
            }
                                             
        packetlen = ip_h->ip_len;  /* Avoid funny length when small (??) */
        INTERPRET(FI_IP_PKTLEN, &ip_h->ip_len) ;     
                     
            /* Compute ptr to IP data */
        ipdp = (union ip_data *) ((u_char *) ip_h + (ip_h->ip_hl<<2 ));
            /* Compute IP option pointer */
        optp = (u_char *) ip_h + sizeof(struct ip);
         
        if (optp < (u_char *) ipdp) {
            while (optp < (u_char *) ipdp) {
                if (*optp == IPOPT_EOL) break;
                else if (*optp == IPOPT_NOP)
                    optp++;  /* just ignore NOP */
                else {
                    INTERPRET(FI_IP_OPTION, optp);
                    optp += *(optp+1);   
                }
            }  /* loop through options */
        }
        else  /* If no options, value is ZERO */
            INTERPRET(FI_IP_OPTION, &zero);
        
        INTERPRET(FI_IP_TOS, &ip_h->ip_tos) ;        
        INTERPRET(FI_IP_PROT, &ip_h->ip_p) ;
        
#ifdef FALIGN
           /* For RISC processors requiring fullword alignment:
            * copy and full-word align src and dest IP addrs */
        bcopy((u_char *)&ip_h->ip_src, ip_addrs.ip_addrs_c, 
                                         sizeof(ip_addrs.ip_addrs_s));
        snet = SADDR & netmask[(SADDR>>30)&3];
        dnet = DADDR & netmask[(DADDR>>30)&3];
#ifdef SUBNETS
        ssbnet = snet;  /* Defaults: subnets same as network numbers. */
        dsbnet = dnet;
        if (SubHdr.Link) {
            register Subnet *sp;
                /* Look up src and dest networks in our table, and if they
                 *  appear as subnetted, extract corresponding subnet address
                 *  using specified mask.
                 */
            if (sp = FindSubnet(snet)) 
                ssbnet = SADDR & sp->Mask;
            if (sp = FindSubnet(dnet))
                dsbnet = DADDR & sp->Mask;
        }
        INTERPRETV(FI_IP_SSBNET, &ssbnet);
        INTERPRETV(FI_IP_DSBNET, &dsbnet);
#endif
        INTERPRETV(FI_IP_SRCNET, &snet) ;            
        INTERPRETV(FI_IP_DSTNET, &dnet) ;                 
        INTERPRET(FI_IP_SRCHST, &SADDR) ;  
        INTERPRET(FI_IP_DSTHST, &DADDR) ;  
#else        
        snet = ip_h->ip_src.s_addr & netmask[(ip_h->ip_src.s_addr>>30)&3] ;
        dnet = ip_h->ip_dst.s_addr & netmask[(ip_h->ip_dst.s_addr>>30)&3] ;
#ifdef SUBNETS
        ssbnet = snet;  /* Defaults: subnets same as network numbers. */
        dsbnet = dnet;
        if (SubHdr.Link) {
            register Subnet *sp;
                /* Look up src and dest networks in our table, and if they
                 *  appear as subnetted, extract corresponding subnet address
                 *  using specified mask.
                 */
            if (sp = FindSubnet(snet)) 
                ssbnet = ip_h->ip_src.s_addr & sp->Mask;
            if (sp = FindSubnet(dnet))
                dsbnet = ip_h->ip_dst.s_addr & sp->Mask;
        }
        INTERPRETV(FI_IP_SSBNET, &ssbnet);
        INTERPRETV(FI_IP_DSBNET, &dsbnet);
#endif
        INTERPRETV(FI_IP_SRCNET, &snet) ;            
        INTERPRETV(FI_IP_DSTNET, &dnet) ;                 
        INTERPRET(FI_IP_SRCHST, &ip_h->ip_src) ;  
        INTERPRET(FI_IP_DSTHST, &ip_h->ip_dst) ;  
#endif
               
            /**** Process higher-level protocol header... *****/
        switch (ip_h->ip_p) {
        
        case IPPROTO_TCP:
            INTERPRET(FI_TCP_SPORT, &ipdp->pk_tcphdr.th_sport) ;
            INTERPRET(FI_TCP_DPORT, &ipdp->pk_tcphdr.th_dport) ;
            return ;
        
        case IPPROTO_UDP:
            INTERPRET(FI_UDP_SPORT, &ipdp->pk_udphdr.uh_sport) ;
            INTERPRET(FI_UDP_DPORT, &ipdp->pk_udphdr.uh_dport) ;
            return ;
        
        case IPPROTO_ICMP:
            INTERPRET(FI_ICMP_TYPE, &ipdp->pk_icmphdr.icmp_type) ;
            return ;        
        
        default:
            return ;        
        } ;
        
        break ;
        
    default:
        return ;
    } ;
    
}


SetEtherType()
    {
    /* Scan through all FPT subtrees that correspond to parsed Ethernet
     *   types, and check if exactly one of them has any invocations,
     *   and if there are no invocations for the Ethernet fields themselves.
     *   If so, can use the Ethernet Type filter of NIT 3.x.  If there are
     *   no invocations at all, do nothing.
     */
     struct FilterTE *ftep;
#ifdef BPF
     struct dlinfo dummy_link;
     struct dlinfo *linkp = &dummy_link;
#endif	/* BPF */
     
     if (Fields[FI_ROOT].field_cuminv == 0 && 
         Fields[FI_PERSEC].field_cuminv == 0) 
           return;
     
     for (ftep = EtherFTable; ftep->FTE_fino; ftep++) {
          if (Fields[ftep->FTE_fino].field_cuminv) {
              break;
          }
     }
     if ((ftep->FTE_fino) && (Fields[ftep->FTE_fino].field_cuminv == 
               Fields[FI_ROOT].field_cuminv ))  {
#ifdef BPF
             if_fd = initdevice(EtherIface, 0, linkp);
#else
             initdevice(ftep->FTE_value);
#endif	/* BPF */
         }
         
     else {
	 /* Else set promiscuous */
#ifdef BPF
	 if_fd = initdevice(EtherIface, 0, linkp);
#else
	 initdevice(-1);
#endif	/* BPF */
     }

#ifdef BPF
     bpf_buf = readinit(if_fd, (struct bpfilter *) 0, &bpf_bufsize);
#endif	/* BPF */
}

#ifdef SUBNETS  
Subnet *FindSubnet(NAddr)
    register u_long NAddr;
    {
    register Subnet *sp;
    for (sp = SubHdr.Link; sp&&(sp->Addr != NAddr); sp= sp->Link);
    return(sp);
}
#endif
