 
/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   * 
 *     as this notice is not removed.                                     *
 **************************************************************************/ 
 

/*
 *   Internet Statistics Facility -- Generic Operation Routines
 *
 *    This module defines a set of operations on a particular class
 *    of Statistical Objects.  In the spirit of object-oriented
 *    programming and information-hiding, these routines are
 *    the only ones which know about the internal structre of the
 *    Statistical Object (SOBJ) data structures.
 *
 */
 
/*
 *
 *          BP Object Class ==   BINARY PACKETS Operation 
 *
 *    An object of the BP class maintains a circular buffer of binary
 *    text  (normally, complete packet headers selected by filter(s)).
 * 
 *    Reading the object returns the saved binary text, starting
 *    from the oldest.  
 *
 *    Operations on BP class objects:
 *
 *       SOBJ *Make_BP( object-name, &Max-number, 1, datalen )
 *
 *       int Write_BP( (SOBJ *) SOBJ-ptr, &value, len)
 *
 *       Read_BP( fd, (SOBJ *) SOBJ-ptr )
 *
 *       Delete_BP( (SOBJ *) SOBJ-ptr )
 *
 *       Clear_BP( (SOBJ *) SOBJ-ptr )
 *
 */
 
#include <stdio.h>
#include <sys/types.h>
#include "stat.h"
#include "sobj.h"



    /* Object area */

#define BPA struct stat_xarea
BPA {
    SOBJ   bpa_sobj ;         /* SOBJ = standard root of area */
    
    short  bpa_pktsize ;      /* Max packet size */
    short  bpa_segsize ;      /* Size of circ buff segment */
    short  bpa_wrapped ;      /* Boolean: true once wraps first time */
    short  bpa_dum1 ;         
    
    BPE   *bpa_nxtslot ;      /* ptr to next slot to be written */
    BPE   *bpa_segp ;         /* Ptr to circular buffer segment */
} ;

struct BP_parms {
    long bpp_maxno ;
} ;

static free_pages() ;
    
SOBJ *Make_BP() ;
boolean  Write_BP(), Read_BP();
void Clear_BP(), Delete_BP() ;

    /* Transfer Vector 
     *   for external access to these routines
     */
GENERICOP BPop = { Make_BP, Write_BP, Read_BP, Clear_BP, Delete_BP } ;

    

SOBJ *Make_BP(objname, parmp, nparms, datalen)
    char *objname;
    struct BP_parms *parmp ;
    int datalen, nparms;
    {
    register BPA *bpap ;
    int vsize = ALIGN(datalen+1) -1 ;   

    if (nparms != 1 || datalen > MAX_BPSIZE) {
        SOBJ_error = "BINPKT: Parameter error";
        return(NULL);
    }
    if ((bpap = (BPA *) malloc( sizeof ( BPA ))) == NULL) {
        SOBJ_error = "BINPKT: Memory overflow" ;
        return(NULL) ;
    }
    bzero(bpap, sizeof(BPA)) ;      
    bpap->bpa_segsize = (sizeof(BPE))*parmp->bpp_maxno ;
    bpap->bpa_pktsize = vsize ;
                                
    return((SOBJ *) bpap) ; 
}    
     
     
int Write_BP( bpap, Valuep, L)
    register BPA *bpap ;
    char *Valuep ;
    int L;
    {
    register BPE *bpep ;
                        
    if (L > bpap->bpa_pktsize) L = bpap->bpa_pktsize; /* 31Dec*/
    
    if (bpap->bpa_segp == NULL) {
        /* Must get circular buffer now... */   
        if ((bpep  = (BPE *) malloc(bpap->bpa_segsize)) == NULL) {
            /* Cannot get new bin page. Count as orphan. */
            bpap->bpa_sobj.sob_orphans++ ;
            return(WRRET_OK) ;
        }
        bpap->bpa_segp = bpap->bpa_nxtslot = bpep ;
    }
    
    bpep = bpap->bpa_nxtslot ;
    bpep->bpe_time = CurrTime ;
    
    bcopy(Valuep, bpep->bpe_packet, bpep->bpe_pktlen = L) ;
    if ((char *) ++(bpap->bpa_nxtslot)  >= 
                   ((char *)bpap->bpa_segp + bpap->bpa_segsize)) {
        bpap->bpa_wrapped = 1 ;
        bpap->bpa_nxtslot = bpap->bpa_segp ;
    }   
    return(WRRET_OK) ;          
}
    
        
void Delete_BP( bpap )
    BPA *bpap ;
    {
    free(bpap->bpa_segp) ;
    free(bpap) ;
}


void Clear_BP( bpap )
    register BPA *bpap ;
    {           
    free(bpap->bpa_segp) ;
    bpap->bpa_segp = bpap->bpa_nxtslot = NULL ; 
    bpap->bpa_wrapped = 0;      
}

    
int Read_BP(fd, bpap)
    int fd ;
    register BPA  *bpap ;
    {
    register BPE *bpep, *last;
    register int length = 0;
    int count;
    int islog = (ISTTY(fd));
    OBJ xobj;
    
    SetupXRep(&xobj, bpap);
    
    last = (bpap->bpa_wrapped)? 
                (BPE *) ((char *) bpap->bpa_segp + bpap->bpa_segsize) :
                bpap->bpa_nxtslot;
    count= last - bpap->bpa_segp;
    
        /* Have to determine total length, first */ 
    if (bpep = bpap->bpa_segp)
        while (bpep < last)  {
            length += ALIGN(bpep->bpe_pktlen);
            bpep++;
        }
        
    if (!PutReadResp(fd, BPr_SIZE(count, length), bpap)) return(FALSE);
    
    if (!ISTTY(fd))
        if (!xdr_int(fd, &count)) 
            return(FALSE);
            
    if (bpap->bpa_wrapped) {
        for (bpep = bpap->bpa_nxtslot ; (char *) bpep < 
                 ((char *) bpap->bpa_segp + bpap->bpa_segsize); bpep++) 
            if (!ReadBP(fd, bpep, &xobj, islog))
                return(FALSE) ;
    }
    if (bpep = bpap->bpa_segp)
        while (bpep < bpap->bpa_nxtslot)  {
            if (!ReadBP(fd, bpep, &xobj, islog))
                return(FALSE) ;
            bpep++; 
        }
    return(TRUE);
    
}   

    
            
