 
/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   * 
 *     as this notice is not removed.                                     *
 **************************************************************************/ 
 

/*
 *   Internet Statistics Facility -- Generic Operation Routines
 *
 *    This module defines a set of operations on a particular class
 *    of Statistical Objects.  In the spirit of object-oriented
 *    programming and information-hiding, these routines are
 *    the only ones which know about the internal structre of the
 *    Statistical Object (SOBJ) data structures.
 *
 */
 
/*
 *
 *  HI Object Class ==  GENERAL HISTOGRAM   
 *
 *    Objects of the HI class build linear frequency distributions.
 *    Specifically, a given unsigned value X is counted in
 *
 *              bin# = floor(X/S) if X <= S*M, 
 *              bin = "default" otherwise.
 *
 *    Here S is a given scale factor, and M is the maximum bin; any values
 *    beyond 1024 are counted in the "default" bin.  If M is unspecified,
 *    M=1024 is assumed. X cannot exceed 4 bytes!
 *
 *
 *    Operations on HI class objects:
 *
 *       SOBJ *Make_HI( objname, &parms, nparms, datalen )
 *
 *                parms = (S, [M])
 *
 *       Write_HI( (SOBJ *) SOBJ-ptr, &value , length)
 *
 *       Read_HI(fd, (SOBJ *) SOBJ-ptr )
 *
 *       Delete_HI( (SOBJ *) SOBJ-ptr )
 *
 *       Clear_HI( (SOBJ *) SOBJ-ptr )
 * 
 *
 *  P2 Object Class ==  POWER-OF-TWO FREQUENCY DISTRIBUTION  
 *
 *    Objects of the P2 class build logarithmic frequency distributions.
 *    Specifically, a given value X is counted in bin log2(X/F), where
 *    F is a given scale factor.  X cannot exceed 4 bytes!
 *
 *
 *    Operations on P2 class objects:
 *
 *       SOBJ *Make_P2( objname, &S, 1, datalen )
 *
 *           S = scale factor.  Bins are:
 *                (0):  value < S
 *                (1):   S <= value <= 2*S-1
 *                (j):   (2**j-1)*S <= value <= (2**j)*S -1
 *                   ...
 *
 *       Write_P2( (SOBJ *) SOBJ-ptr, &value, length)
 *
 *       Read_P2(fd, (SOBJ *) SOBJ-ptr )
 *
 *       Delete_P2( (SOBJ *) SOBJ-ptr ) 
 *
 *       Clear_P2( (SOBJ *) SOBJ-ptr )  
 *
 */
 
#define MAXHBINS    1024 

#include <stdio.h>
#include <sys/types.h>
#include "stat.h"
#include "sobj.h"
    
SOBJ *Make_HI(), *Make_P2(), *make_hist() ;
boolean  Write_HI(), Read_HI(), Write_P2(), Read_P2() ;
void Clear_HI(), Delete_HI() ;

        /* Transfer Vector 
         * for external access to these routines
         */
GENERICOP HIop = { Make_HI, Write_HI, Read_HI, Clear_HI, Delete_HI } ;
GENERICOP P2op = { Make_P2, Write_P2, Read_P2, Clear_HI, Delete_HI } ;
    
    /*
     *  Define HI/P2 object data structures
     *
     */
#define HistA struct stat_freq_area
#define esize sizeof(HistE)
 
HistA {     /* Histogram Object area */
    SOBJ   hia_sobj ;         /* SOBJ = standard root of area */
    
    short  hia_bsize ;        /* Size of bin area (bytes) */
    short  hia_nbins ;        /* Size of bin area (#elements) */    
    
    long   hia_sum1 ;         /* sum / 2**16 */ 
    long   hia_sum2 ;         /* sum mod 2**16  */  
    HistDATA hia_data;        /* scale factor, avg, min, max */
    HistE   *hia_binp;          /* pointer to bin vector */
} ;

#define   hia_totalno  hia_sobj.sob_totalno
#define   hia_scalef   hia_data.Hist_scalef    /* Scalefactor */    
#define   hia_avg      hia_data.Hist_avg       /* Current average value */
#define   hia_minval   hia_data.Hist_min       /* Smallest observed value */
#define   hia_maxval   hia_data.Hist_max       /* Largest observed value */
#define   hia_default  hia_data.Hist_default   /* Default count */

struct Hist_parms {
    u_long  hip_scalef;
    u_long  hip_maxbin;
} ;


SOBJ *Make_HI(objname, parmp, nparms, datalen)
    char *objname;
    struct Hist_parms *parmp ;
    int   datalen, nparms;
    {
    int nbins = MAXHBINS;
        
    if (nparms < 1) {
        SOBJ_error = "HI: parm missing";
        return(NULL);
    }       
    if (nparms > 1) {
        if (parmp->hip_maxbin > MAXHBINS-1) {
            SOBJ_error = "HI: too many bins";
            return(NULL);
        }
        nbins = parmp->hip_maxbin+1;
    }
    if (datalen  > 4) {
        SOBJ_error = "HI: cannot handle > 4bytes";
        return(NULL);
    }                
    return((SOBJ *) make_hist(parmp->hip_scalef, nbins)) ;      
}    


SOBJ *Make_P2(objname, parmp, nparms, datalen)
    char *objname;
    struct Hist_parms *parmp ;
    int datalen, nparms;
    {       
    if (nparms < 1) {
        SOBJ_error = "P2: parm missing";
        return(NULL);
    }   
    if (datalen  > 4) {
        SOBJ_error = "HI: cannot handle > 4bytes";
        return(NULL);
    }   
    return((SOBJ *) make_hist(parmp->hip_scalef, 8*datalen)) ;      
}    

    /* 
     *   Common object Make routine...
     */
SOBJ *make_hist(scalef, nbins)
    int   scalef, nbins;
    {
    register HistA *hiap ;
    
    if ((hiap = (HistA *) malloc(sizeof (HistA))) == NULL) {
        SOBJ_error = "HI/P2: Mem full" ;
        return(NULL) ;
    }       
    
    hiap->hia_nbins = nbins;
    hiap->hia_scalef = (scalef)?scalef:1 ;
    hiap->hia_bsize =  (esize)*hiap->hia_nbins;

    if ((hiap->hia_binp = (HistE *) malloc(hiap->hia_bsize))== NULL) {
        SOBJ_error = "HI/P2: Mem Full" ;
        return(NULL) ;
    }
    Clear_HI(hiap);
    return((SOBJ *) hiap) ;  /* SUCCESS */      
}    
     
     
boolean Write_HI( hiap, Valuep, L)
    register HistA *hiap ;
    char *Valuep ;
    int L;
    {
    register u_long Value = LFETCH(Valuep, L) ;
    
    if (Value < hiap->hia_minval) hiap->hia_minval = Value ;
    if (Value > hiap->hia_maxval) hiap->hia_maxval = Value ;
    
    hiap->hia_sum2 += Value ;
    hiap->hia_sum1 += (hiap->hia_sum2 >> 16) ;
    hiap->hia_sum2 &= 0xffff ;
                              
    if ((Value /= hiap->hia_scalef) >= hiap->hia_nbins)
        hiap->hia_default++;
    else        
        hiap->hia_binp[Value]++ ;
    return(0) ;
}
    
     
boolean Write_P2( hiap, Valuep, L)
    register HistA *hiap ;
    char *Valuep ;
    int L;
    {
    register u_long Value = LFETCH(Valuep, L) ;
    register int i = 0 ;
                              
    if (Value < hiap->hia_minval) hiap->hia_minval = Value ;
    if (Value > hiap->hia_maxval) hiap->hia_maxval = Value ;
    
    hiap->hia_sum2 += Value ;
    hiap->hia_sum1 += (hiap->hia_sum2 >> 16) ;
    hiap->hia_sum2 &= 0xffff ;
    
    Value /= hiap->hia_scalef ;
    
    while (Value) {i += 1 ; Value >>= 1 ;}  
        /* i is number significant bits in Value */     
    hiap->hia_binp[i]++ ;   
    return(0) ;
}
            
void Delete_HI( hiap )
    HistA *hiap ;
    {
    if (hiap->hia_binp) free(hiap->hia_binp);
    free(hiap) ;
}


void Clear_HI( hiap )
    HistA *hiap ;
    {
    bzero( (char *) hiap->hia_binp, hiap->hia_bsize );  
    hiap->hia_minval = INFINITY;
    hiap->hia_maxval = hiap->hia_default = 0;
    hiap->hia_sum1 = hiap->hia_sum2 = 0;
}


boolean Read_HI(fd, hiap)
    int fd;
    register HistA  *hiap ;
    {
    register HistE *tbp = &hiap->hia_binp[hiap->hia_nbins] ;
    register int left = hiap->hia_nbins;
    int i;
    HistE *hep = hiap->hia_binp;
    u_long Avg();
            
    while (left > 0) 
        if (*--tbp) break; else left--;
       /* Scan backwards from end for first non-zero count */      
    if (left<(hiap->hia_nbins-1)) left++ ;  /* Include one zero bin */
     
    hiap->hia_avg = Avg(hiap->hia_sum1, hiap->hia_sum2, hiap->hia_totalno); 
    
    i = left;
    if (!PutReadResp(fd, HIr_SIZE(left), hiap) ||
        !ReadHI(fd, &hep, &i, &hiap->hia_data, ISTTY(fd))
        )
        return(FALSE); 
    return(TRUE); 
}


boolean Read_P2(fd, hiap)
    int fd;
    register HistA  *hiap ;
    {
    register HistE *tbp = &hiap->hia_binp[hiap->hia_nbins] ;
    register int left = hiap->hia_nbins;
    int i;
    HistE *hep = hiap->hia_binp;
    u_long Avg();
            
    while (left > 0) 
        if (*--tbp) break; else left--;
       /* Scan backwards from end for first non-zero count */      
    if (left<(hiap->hia_nbins-1)) left++ ;  /* Include one zero bin */
     
    hiap->hia_avg = Avg(hiap->hia_sum1, hiap->hia_sum2, hiap->hia_totalno); 
    
    i = left;
    if (!PutReadResp(fd, P2r_SIZE(left), hiap) ||
        !ReadP2(fd, &hep, &i, &hiap->hia_data, ISTTY(fd))
        )
        return(FALSE);
    return(TRUE);  
}


u_long Avg(sum1, sum2, totalno)
    u_long sum1, sum2, totalno;
    {
    float SUM;
        
    if (totalno) {
        SUM = ((float) sum1)*65536 + sum2;
        return(SUM/totalno);
    } 
    else
        return(0);
}
    
