 
/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   * 
 *     as this notice is not removed.                                     *
 **************************************************************************/ 
 
/*
 *   Internet Statistics Facility -- Generic Operation Routines
 *
 *    This module defines a set of operations on a particular class
 *    of Statistical Objects.  In the spirit of object-oriented
 *    programming and information-hiding, these routines are
 *    the only ones which know about the internal structre of the
 *    Statistical Object (SOBJ) data structures.
 *
 */
 
/* CHANGES:
 *   18Nov88 RTBraden -- Fix bug in read: missing SetupXRep() call.
 *
 */
 
/*
 *
 *          WS Object Class ==  working-set
 *
 *    Objects of the WS class compute a frequency distribution related to
 *    the degree of locality of reference in a sequence of field values.
 *    This problem is related closely to virtual memory algorithms.  
 *
 *    Let C(n) be the number of number of values in the given sequence
 *    that fall outside a set of the n most recent ones.  If there are N
 *    values in the sequence, C(n)/N is the corresponding probability
 *    ("page fault probability").  The WS object returns C(1), C(2),
 *    C(4), C(8), ... C(4096).
 *
 *    Operations on WS class objects:
 *
 *       SOBJ *Make_WS( object-name, &parmlist, nparms, datalen )
 *
 *       Write_WS( (SOBJ *) SOBJ-ptr, &value, len)
 *
 *       Read_WS( fd, (SOBJ *) SOBJ-ptr )
 *
 *       Delete_WS( (SOBJ *) SOBJ-ptr )
 *
 *       Clear_WS( (SOBJ *) SOBJ-ptr )
 *
 */
 
#include <stdio.h>
#include <sys/types.h>
#include "stat.h"
#include "sobj.h"

    /*     
     *  Define WS object data structures
     *
     */  
     
#define MAX_PAGES  8
    
#define BIN_pwr2  8   /* log2( number bins per extension page )*/
#define BIN_PAGE  (1<<BIN_pwr2)   /* Number bins per extension page */
#define BIN_MASK  BIN_PAGE -1
#define BIN_MAX   BIN_PAGE*MAX_PAGES

        /* Map bin index into bin address */    
#define BIN_ADDR(wsap, i) (WSE *) ( \
             (char *) wsap->wsa_page[((i)>>BIN_pwr2)]+ \
             wsap->wsa_esize*((i)&BIN_MASK) ) 
     
    /* Bin format */
    
#define WSE struct ws_bin_element
WSE {
    struct FBlinks wse_fb;  /* forward and reverse links */ 
    char   wse_value[MAX_DLENG];
            /* Actually reserve 4*((dleng+3)/4) bytes */
};


    /* Object area format */
    
#define WSA struct stat_xarea
WSA {
    SOBJ   wsa_sobj ;         /* SOBJ = standard root of area */
    
    int    wsa_esize ;            /* Size of element */ 
    int    wsa_nextbin ;          /* Index of first available bin */
    
    struct FBlinks wsa_head;      /* Head of LRU chain of bins */
    WSE   *wsa_page[MAX_PAGES] ;  /* Ptrs to bin segments */
    /*----------------------------------------------------*/
    int    wsa_count[MAX_LOG2N+1];   /* Counts */
} ;


SOBJ *Make_WS() ;
boolean Write_WS(), Read_WS() ;
void Clear_WS(), Delete_WS() ;

    /* Transfer Vector 
     *   for external access to these routines
     */
GENERICOP WSop = { Make_WS, Write_WS, Read_WS, Clear_WS, Delete_WS } ;

    

SOBJ *Make_WS(objname, parmp, nparms, datalen)
    char *objname;
    char *parmp;
    int datalen, nparms;
    {
    register WSA *wsap ;

    if (datalen > MAX_DLENG){
        SOBJ_error = "WS: Field size too big";
        return(NULL);
    }
    if ((wsap = (WSA *) malloc(sizeof(WSA))) == NULL) {
        SOBJ_error = "WS: Mem overflow" ;
        return(NULL) ;
    }
    bzero(wsap, sizeof(WSA)) ;  /* Clear entire area... */      
    wsap->wsa_esize = sizeof(WSE) - MAX_DLENG + ALIGN(datalen);
    wsap->wsa_nextbin = 0 ;     
                                
    return((SOBJ *) wsap) ;  /* SUCCESS!! */
}    
     
     
boolean Write_WS( wsap, Valuep, L)
    register WSA *wsap ;
    char *Valuep ;
    int L; 
    {
    register WSE *wsep;
    register int K, i;
     
    wsep = (WSE *) wsap->wsa_head.FB_Fp;
    K = 0;
    while (wsep) {
        if (!bcmp(wsep->wse_value, Valuep, L)) {
            /* Found existing bin.  If it is first, it would be included
             *   in ANY set, so and it can remain first in LRU list, so we
             *   are all done.  Else remove it from list, so it can be
             *   moved to front of list later.
             */
            if (K == 0) return(WRRET_OK);
            FBrem(&wsep->wse_fb);
            break;
        }
        K++;
        wsep = (WSE *) wsep->wse_fb.FB_Fp ;
    }
    if (wsep == NULL) {
        /* No matching bin... Must create new one.       */      
        if (wsap->wsa_nextbin >= BIN_MAX ) {
            /* We have run out of new bins. Count as orphan. */
            ((SOBJ *) wsap)->sob_orphans++ ;
            return(WRRET_OK) ;
        }
        
        if ((wsep = BIN_ADDR( wsap, wsap->wsa_nextbin)) == NULL) {
            /*  must obtain new page */
            wsep  = (WSE *) malloc(wsap->wsa_esize * BIN_PAGE) ;
            if (wsep == NULL) {
                /* Cannot get new bin page. Count as orphan. */
                ((SOBJ *) wsap)->sob_orphans++ ;
                return(WRRET_OK) ;
            }
            wsap->wsa_page[(wsap->wsa_nextbin >> BIN_pwr2)] = wsep ;
        }
                
        bcopy(Valuep, wsep->wse_value, L);
        wsep->wse_fb.FB_Fp = wsep->wse_fb.FB_Bp = 0;
        wsap->wsa_nextbin++;
    }
    
        /* At this point, wsep is newly-referenced bin, and K is its
         *    position in the LRU list. Compute i = floor(log2(K)),
         *    and increment count[i] by one.
         */           
    
    i = -1;
    while (K) {
        i++; K >>= 1;
    }
    if (i >= 0) {
        if (i <= MAX_LOG2N)
            wsap->wsa_count[i]++;
        else
            ((SOBJ *) wsap)->sob_orphans++ ;
    }   
    FBins(&wsep->wse_fb, &wsap->wsa_head);
    return(WRRET_OK) ;  
}
    
        
void Delete_WS( wsap )
    WSA *wsap ;
    {
    WSfree_pages(wsap) ;
    free(wsap) ;
}


void Clear_WS( wsap )
    register WSA *wsap ;
    {           
    wsap->wsa_nextbin = 0;
    wsap->wsa_head.FB_Fp = wsap->wsa_head.FB_Bp = 0;
    bzero(wsap->wsa_count, (1+MAX_LOG2N)*sizeof(int));
    WSfree_pages(wsap) ;                
}

        
WSfree_pages( wsap )
    register WSA *wsap ;
    {
    register int i ;
    register WSE *wsep ;
    
    for (i = 0;i<MAX_PAGES;i++) {
        if (wsep = wsap->wsa_page[i]) {
            free(wsep) ;
            wsap->wsa_page[i] = NULL ;
        }
        else break ;
    }
}
    
#define OUTSIZE 256 

boolean Read_WS(fd, wsap)
    int fd;
    register WSA  *wsap ;
    {
    register int i;
    register int sum;
    u_long  Ans[1+MAX_LOG2N], *vecptr = Ans;
    int  veclen;
    OBJ  xobj;  
    
    for (i = MAX_LOG2N; i >= 0; i--) if (wsap->wsa_count[i]) break;
    veclen = (i<MAX_LOG2N)? i+2:i+1;
    
    if (!PutReadResp(fd, WSr_SIZE(veclen), wsap))
        return(FALSE);
             
        /* Must propagate counts to front... add them all up, and
         *  then subtract off each successive count value...
         */
    sum = 0;
    for (i=0; i < veclen; i++) sum += wsap->wsa_count[i];   
    for (i=0; i < veclen; i++) {
        Ans[i] = sum;
        sum -= wsap->wsa_count[i]; 
    } 
    SetupXRep(&xobj, wsap);   
    if (!ReadWS(fd, &vecptr, &veclen, &wsap->wsa_nextbin, &xobj, ISTTY(fd)))
        return(FALSE);
    
    return(TRUE);
}   

